/***
* DBClassy.ch
*
* Class(y) version of TBrowse sub class to allow database browsing
* with:
*   For scope
*   While scope
*   Append mode
*/

#include "Class(y).ch"

CREATE CLASS TBfw FROM TBrowse

HIDDEN:
  VAR goNext
  VAR goPrev

  METHOD TBPrev
  METHOD TBNext

EXPORT:
  VAR while        // The WHILE scope
  VAR for          // The FOR scope
  VAR appendMode   // Is the user allowed to append records?

  VAR goFirst      // Code block to move to the first record
  VAR goLast       // Code block to move to the last record

  METHOD TBFirst
  METHOD TBLast
  METHOD DBSkipBlock

  METHOD Init

END CLASS


FUNCTION TBrowseFW(bWhile, bFor, bFirst, bLast)

RETURN TBfw():New(bWhile, bFor, bFirst, bLast)


METHOD Init(bWhile, bFor, bFirst, bLast), ()

  self:while   := bWhile
  self:for     := bFor
  self:goFirst := bFirst
  self:goLast  := bLast

  self:goNext     := {|| self:TBNext() }
  self:goPrev     := {|| self:TBPrev() }
  self:appendMode := .F.

  self:goTopBlock    := {||  self:TBFirst() }
  self:goBottomBlock := {||  self:TBLast()  }
  self:skipBlock     := {|n| self:DbSkipBlock(n) }

  self:goTop()

RETURN Self


	
// Routine called by goTopBlock. It must move to the first record
// in the database that matches both the FOR and WHILE scopes.
// These are stored as code blocks in the TBrowse object.
// If the routine does not find a matching record, it must
// position the record pointer at end of file.

METHOD FUNCTION TBFirst()

  // This takes us to the first one
  Eval(self:goFirst)

  // But we must ensure this also matches the FOR scope,
  // without moving outside the records matching the WHILE scope.
  DO WHILE !Eof() .AND. Eval(self:while) .AND. !Eval(self:for)
    SKIP
  ENDDO

  // If we are not on a valid record, move to end of file.
  IF Eof() .OR. !Eval(self:while)
    GOTO 0
  ENDIF

RETURN NIL


// Routine called by goBottomBlock to move to the last record
// in the database matching the for and while condition, and
// taking append mode into account.

METHOD FUNCTION TBLast()

  // If append mode is on, move to end of file
  IF self:appendMode
    GOTO 0
  ELSE
    // This takes us to the last record
    Eval(self:goLast)

    // This ensures the record matches both the for and while scope
    DO WHILE !Bof() .AND. Eval(self:while) .AND. !Eval(self:for)
      SKIP -1
    ENDDO

    // If we are not on a valid record, go to end of file
    IF Bof() .OR. !Eval(self:while)
      GOTO 0
    ENDIF
  ENDIF

RETURN NIL


// Move to next record supporting for and while conditions
// and append mode. Returns .T. if successful, .F. otherwise
// If it can't move, it doesn't change the record pointer.

METHOD FUNCTION TBNext()

// First save record pointer so we can restore it
// if we can't move.
LOCAL nSaveRecNum := Recno()
LOCAL lMoved := .T.

  IF Eof()
    lMoved := .F.
  ELSE
    SKIP
    DO WHILE !Eval(self:for) .AND. Eval(self:while) .AND. !Eof()
      SKIP
    ENDDO

    IF Eof() .AND. self:appendMode
      // fine ...
    ELSEIF !Eval(self:while) .OR. Eof()
      IF !Eval(self:while) .AND. self:appendMode
        GOTO 0
      ELSE
        lMoved := .F.
        GOTO nSaveRecNum
      ENDIF
    ENDIF
  ENDIF

RETURN lMoved


// Move to previous record supporting for and while conditions
// and append mode

METHOD FUNCTION TBPrev()

LOCAL nSaveRecNum := Recno()
LOCAL lMoved := .T.

  // Take special care at Eof() - a SKIP -1 is not sufficient
  // because of the filters
  IF Eof()
    Eval(self:goLast)
  ELSE
    SKIP -1
  ENDIF

  DO WHILE !Eval(self:for) .AND. Eval(self:while) .AND. !Bof()
    SKIP -1
  ENDDO

  IF !Eval(self:while) .OR. Bof()
    GOTO nSaveRecNum
    lMoved := .F.
  ENDIF

RETURN lMoved



// Skipblock routine for database browses. Needs supporting
// code blocks that move forward by one record and move backward
// by one record. They must both return logicals which indicate
// whether they suceeded or not.
// This function returns how many records it skipped, as required
// by TBrowse:skipBlock.

METHOD FUNCTION DbSkipBlock(n)

LOCAL nSkipped := 0

  // Idiosyncrasy for database skipper
  IF n = 0
    SKIP 0
  ELSEIF n > 0
    DO WHILE nSkipped != n .AND. Eval(self:goNext)
      nSkipped++
    ENDDO
  ELSE
    DO WHILE nSkipped != n .AND. Eval(self:goPrev)
      nSkipped--
    ENDDO
  ENDIF

RETURN nSkipped
